function [T,V,E,TE] = boxTetraMesh(Xno,Yno,Zno)
% Generate the tetrahedron mesh of a box with given grid numbers
%
% Input  --
%    Xno :: number of nodes along the x-axis
%    Yno :: number of nodes along the y-axis
%    Zno :: number of nodes along the z-axis
%
% Output --
%    T   :: Tno x 4 matrix, the index list of all tetrahedrons,
%           Tno = #(tetrahedrons) = 6 * (Xno-1) * (Yno-1) * (Zno-1)
%    V   :: Vno x 3 matrix, the 3-D coordinates of each node point
%           Vno = #(node points) = Xno * Yno * Zno
%    E   :: Eno x 2 matrix, the index list of all (ordered) edges
%           Eno = #(edges) = ( 7*Xno*Yno - 4*(Xno + Yno) + 2 ) * (Zno - 1) 
%                            + (3*Xno - 2) * (Yno - 1) + (Xno - 1)
%    TE  :: Tno x 2 matrix, the edge indices on each tetrahedron
%
% By Wei-Qiang Huang (13.04.11-q1.0, revised: 13.04.15-q4.0)

Vno = Xno * Yno * Zno;
Bno = (Xno-1) * (Yno-1) * (Zno-1);

% BEGIN T :: Generate the tetrahedral list T with size Tno x 4 ------------
%
T = T_blkshift(Xno, Yno, Zno);
%
% ---------------------------------------------------------------- :: END T

% BEGIN E :: Generate the edge list E with size Eno x 2 -------------------
%
E = E_blkshift(Xno, Yno, Zno);
%         ---------- Only for checking the beta version ----------
check = 0;
if check == 1
    % Ei and Ej denote the vertex indices of the individual edge from the 
    % vertex i to vertex j with the order given by the columns Ei and Ej:
    % Ei = [V1  V2  V3  V1  V2  V1]    to    Ej = [V2  V3  V4  V3  V4  V4]
    Ei = T;
    Ej = circshift(T, [0 -1]);
    Ei(:, 4:6) = [T(:, 1:2), T(:, 1)];
    Ej(:, 4:6) = [T(:, 3:4), T(:, 4)];
    % Ge is the adjacency matrix of the edge connections. We swap Ei and Ej
    % in advanceIn to prevent the additional transpose operation and obtain
    % an vertex indices of the edge list E so that E(:, 1) and E(:, 3) are, 
    % resp., in the increasing order of the vertex indices.
    Ge = sparse(Ej, Ei, 1, Vno, Vno);
    [E(:,2), E(:,1)] = find( Ge );
    clear Ei Ej Ge
end
%         ---------- Only for checking the beta version ----------
%
% ---------------------------------------------------------------- :: END E

% BEGIN TE :: Generate the edge indices for each tetrahedron, which is
%             called TE with size Tno x 6.
%
% This subroutin has two methods: (1)block-shifted; and (2) graphical
% (1) The 'block-shifted' technique applies the block structure together 
%     with the vetorization skill to provide a fast method to construct
%     the tetra-edge connection matrix TE.
% (2) The 'graphical' computing for the TE matrix uses the graph relations
%     between tetrahedrons and edges. Compare with (1), it provides an 
%     intuitive and memory-saving way to build the TE matrix.
%
if min([Xno, Yno, Zno]) > 10;  
    % block-shifted generation
    TE = TE_blkshift(Xno, Yno, Zno);
else
    % robust generation
    Tno  = 6*Bno;
    Eno  = size(E, 1);
    Gev = sparse(repmat(1:Eno, 1, 2), E, 1, Eno, Vno);
    Gvt = sparse(T, repmat(1:Tno, 1, 4), 1, Vno, Tno);
    Get = Gev * Gvt;
    [TE, ~] = find(Get == 2);
    TE = ( reshape(TE, 6, Tno) )';
    clear Gev Gvt Get
end
%
% --------------------------------------------------------------- :: END TE

% BEING V :: Generate the vertex coordinates V with size Vno x 3 ----------
%
x = linspace(0, 1, Xno);
y = linspace(0, 1, Yno);
z = linspace(0, 1, Zno);
[y, x, z] = meshgrid(y, x, z);
V = [x(:), y(:), z(:)];
%
% ---------------------------------------------------------------- :: END V
end


% BEGIN SUBROUTINE :: T_blkshift ------------------------------------------
%
function T = T_blkshift(Xno, Yno, Zno)

Bno = (Xno-1) * (Yno-1) * (Zno-1);

% Step 1: Build the unit tetrahedral mesh for the first block whose indices
% are             floor 1:                       floor 2:
%            [ 1+Xno    2+Xno ]        [ XYno+1+Xno,   XYno+2+Xno ]
%            [ 1        2     ]        [ XYno+1        XYno+2     ]
% where XYno = Xno * Yno.
t(:, :, 1) = [1, 2; Xno+1, Xno+2];
t(:, :, 2) = t(:, :, 1) + Xno*Yno;
B = zeros(6, 4);
B(:, 1) = t(1,1,1);
B(:, 2) = [t(1,2,1); t(1,2,1); t(2,1,1); t(2,1,1); t(1,1,2); t(1,1,2)];
B(:, 3) = [t(2,2,1); t(1,2,2); t(2,2,1); t(2,1,2); t(1,2,2); t(2,1,2)];
B(:, 4) = t(2,2,2);

% Step 2: Generate the proto-list of tetrahedral indices according to the
%         basic tetrahedral block B
T = repmat(B, Bno, 1);

% Step 3: Determine the increments of index for each block on the floor 1, 
%         which has (Xno-1) x (Yno-1) blocks.
indshift = repmat(0:Xno-2, 1, Yno-1);
temp     = Xno : Xno : Xno*(Yno-2);
temp     = repmat(temp, Xno-1, 1);
temp     = ( temp(:) )';

indshift(Xno:end) = indshift(Xno:end) + temp;

% Step 4: Determine the increments of index for each block on each floor,
%         totally with Bno blocks.
indshift = repmat(indshift, 1, Zno-1);
XYno     = Xno * Yno;
Xm1Ym1no = (Xno-1) * (Yno-1);
temp     = XYno : XYno : XYno*(Zno-2);
temp     = repmat(temp, Xm1Ym1no, 1);
temp     = ( temp(:) )';

indshift( Xm1Ym1no+1 : end ) = indshift( Xm1Ym1no+1 : end ) + temp;

clear temp

% Step 5: Add the increment of index to the tetrahedral list to complete
%         the tetrahedral list T
indshift = indshift(2:end);
indshift = repmat(indshift, 6, 1);
indshift = indshift(:);
indshift = repmat(indshift, 1, 4);

T(7:end, :) = T(7:end, :) + indshift;
end
%
% -------------------------------------------- END SUBROUTINE :: T_blkshift

% BEGIN SUBROUTINE :: E_blkshift ------------------------------------------
%
function E = E_blkshift(Xno, Yno, Zno)

Xm1  = Xno - 1;
XYno = Xno*Yno;

eX    = 7*Xno - 4;
eXxYm1 = eX*(Yno - 1);
eXxY   = eXxYm1 + 3*Xno - 2;

E(:, 1)     = reshape(repmat(1:Xno-1, 7, 1), [], 1);
E(1:7,   2) = [2; Xno+1; Xno+2; XYno+1; XYno+2; XYno+Xno+1; XYno+Xno+2];
E(8:end, 2) = E(1:7*(Xno-2), 1) + repmat(E(1:7, 2), Xno-2, 1);

blkE = [ Xno   Xno + Xno; 
         Xno   Xno + XYno; 
         Xno   Xno + XYno + Xno ];
E = [E; blkE];

s = eX + 1;
t = eXxYm1;
E(s:t, :) = reshape(repmat(Xno:Xno:Xno*(Yno-2), eX, 2), [], 2);
E(s:t, :) = E(s:t, :) + repmat(E(1:eX, :), Yno-2, 1);

s = eXxYm1 + 1;
t = eXxY - 1;
E(s:t, :) = reshape(repmat(E(end, 1) + (1:Xm1), 3, 2), [], 2);
E(s:t, 2) = E(s:t, 2) + repmat([1; XYno; XYno+1], Xm1, 1);
E = [ E; [XYno  2*XYno] ];

s = eXxY + 1;
t = eXxY*(Zno-1);
E(s:t, :) = reshape(repmat(XYno:XYno:XYno*(Zno-2), eXxY, 2), [], 2);
E(s:t, :) = E(s:t, :) + repmat(E(1:eXxY, :), Zno-2, 1);


%
s  = t + 1;
t  = s + 3*(Xno-1) - 1;
e0 = E(end, 1);
E(s:t, :) = reshape(repmat( (e0 + 1 : e0 + Xm1), 3, 2), [], 2);
E(s:t, 2) = E(s:t, 2) + repmat([1; Xno; Xno+1], Xm1, 1);
E = [ E; [E(end-2, 2)  E(end, 2)] ];

E = [E; repmat(E(s:t+1, :), Yno-2, 1)];
eZX = 3*(Xno - 1) + 1;
s  = t + 2;
t  = s + eZX * (Yno-2) - 1;
E(s:t, :) = E(s:t,:) + reshape(repmat(Xno:Xno:Xno*(Yno-2), eZX, 2), [], 2);

e0 = E(end, 1) + 1;
E = [E; [ (e0 : e0 + Xno - 2); (e0 + 1 : e0 + Xno - 1) ]' ];
end
%
% -------------------------------------------- END SUBROUTINE :: E_blkshift

% BEGIN SUBROUTINE :: TE_blkshift -----------------------------------------
%
function TE = TE_blkshift(Xno, Yno, Zno)
x1 = 7;
y1 = 7*Xno - 4;
z1 = 7 * (Xno-1)*(Yno-1) + 3 * (Xno+Yno) - 5;

% Step 1: x-direction shift on the first row.
%         Tetra-edge connections of the first row block without the tails, 
%         size(TE) = 6*(Xno-2)
g    = 1 + y1;
h    = 1 + z1;
blkE = ...
[ 1    3    7    9       13        ( g + x1 ) + 3
  1    5    7   11       13        ( h + x1 ) + 1
  2    3    7    g        g + 4    ( g + x1 ) + 3
  2    6    7    g + 3    g + 4      g + z1
  4    5    7    h        h + 2    ( h + x1 ) + 1
  4    6    7    h + 1    h + 2      g + z1       ];

indshift     = 7 : 7 : ( 7*(Xno - 3) );
indshift     = repmat(indshift, 6, 1);
indshift     = indshift(:);

TE           = repmat(blkE, Xno-2, 1);
TE(7:end, :) = TE(7:end, :) + repmat(indshift, 1, 6);

% Step 2: Complete the x-direction shift on the first row.
%         Tetra-edge connections of the last block of the first row.
%         size(TE) = 6*(Xno-1)
e0   = 7 * (Xno - 2) + 1;
f    = e0 + x1;
g    = e0 + y1;
h    = e0 + z1;
blkE = ...
[ e0        e0 + 2    e0 + 6    f        f + 2  ( g + x1 ) + 1
  e0        e0 + 4    e0 + 6    f + 1    f + 2    h + x1
  e0 + 1    e0 + 2    e0 + 6    g        g + 4  ( g + x1 ) + 1
  e0 + 1    e0 + 5    e0 + 6    g + 3    g + 4    g + z1
  e0 + 3    e0 + 4    e0 + 6    h        h + 2    h + x1
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2    g + z1       ];

TE = [TE; blkE];

% Step 3: y-direction shift on the first floor.
%         Tetra-edge connections of the first floor without the last row.
%         size(TE) = 6*(Xno-1)*(Yno-2)
x_blks  = 6 * (Xno-1);
indshift = y1 : y1 : y1*(Yno-3);
indshift = repmat(indshift, x_blks, 1);
indshift = indshift(:);

TE                  = repmat(TE, Yno-2, 1);
TE(x_blks+1:end, :) = TE(x_blks+1:end, :) + repmat(indshift, 1, 6);

clear indshift

% Step 4: x-direction shift of the last row on the first floor.
%         Tetra-edge connections of the first floor without the last block. 
%         size(TE) = 6*(Xno-1)*(Yno-2) + 6*(Xno-2)
e0 = (7*(Xno - 1)+3)*(Yno - 2) + 1;
x1_3 = 3;
f    = e0 + x1;
g    = e0 + y1;
h    = e0 + z1;
blkE = ...
[ e0        e0 + 2    e0 + 6    f + 1    f + 5   ( g + x1_3 ) + 1
  e0        e0 + 4    e0 + 6    f + 3    f + 5   ( h + x1 ) + 1
  e0 + 1    e0 + 2    e0 + 6    g        g + 2   ( g + x1_3 ) + 1
  e0 + 1    e0 + 5    e0 + 6    g + 1    g + 2     g + z1
  e0 + 3    e0 + 4    e0 + 6    h        h + 2   ( h + x1 ) + 1
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2     g + z1         ];

corval   = reshape(repmat( (2:Xno-3), 6, 1), [], 1);
indshift = 7*ones(6,6);
indshift(3:4, 4:5) = 3;
indshift([1,3,4,6], 6) = 3;

TE = [TE; zeros(6, 6); repmat(indshift, Xno-3, 1)];
s = 6*(Xno-1)*(Yno-2);
TE(s+13:end, :) = TE(s+13:end, :) .* repmat(corval, 1, 6);
TE(s+1:end, :) = TE(s+1:end, :) + repmat(blkE, Xno-2, 1);

clear indshift

% Step 5: Complete the first floor.
%         Tetra-edge connections of the last one block on the first floor.
%         size(TE) = 6*(Xno-1)*(Yno-1)
e0   = (7*(Xno - 1)+3)*(Yno - 2) + 7 * (Xno - 2) + 1;
f    = e0 + x1;
g    = blkE(3, 4) + 3*(Xno-2);
h    = e0 + z1;
blkE = ...
[ e0        e0 + 2    e0 + 6    f        f + 2    g + 3
  e0        e0 + 4    e0 + 6    f + 1    f + 2    h + x1
  e0 + 1    e0 + 2    e0 + 6    g        g + 2    g + 3
  e0 + 1    e0 + 5    e0 + 6    g + 1    g + 2    g + z1
  e0 + 3    e0 + 4    e0 + 6    h        h + 2    h + x1
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2    g + z1 ];

TE = [TE; blkE];

% Step 6: z-direction shift.
%         Tetra-edge connections without the last floor.
%         size(TE) = 6*(Xno-1)*(Yno-1)*(Zno-2)
xy_blks  = x_blks * (Yno-1);
indshift = z1 : z1 : z1*(Zno-3);
indshift = repmat(indshift, xy_blks, 1);
indshift = indshift(:);

TE                  = repmat(TE, Zno-2, 1);
TE(xy_blks+1:end, :) = TE(xy_blks+1:end, :) + repmat(indshift, 1, 6);

clear indshift

% Step 7: x-direction shift of the first row on the topest floor without 
%         the last block
%        size(TE) = 
%
e0   = z1*(Zno - 2) + 1;
y1_3 = 3*Xno - 2;
f    = e0 + x1;
g    = e0 + y1;
h    = e0 + z1;
blkE = ...
[ e0        e0 + 2    e0 + 6    f + 1    f + 5    ( g + x1 ) + 3
  e0        e0 + 4    e0 + 6    f + 3    f + 5    ( h + x1_3 ) + 1
  e0 + 1    e0 + 2    e0 + 6    g        g + 4    ( g + x1 ) + 3
  e0 + 1    e0 + 5    e0 + 6    g + 3    g + 4    (e0 + y1_3) + z1
  e0 + 3    e0 + 4    e0 + 6    h        h + 2    ( h + x1_3 ) + 1
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2    (e0 + y1_3) + z1 ];

% corval   = reshape(repmat( (2:Xno-3), 6, 1), [], 1);
indshift = 7*ones(6,6);
indshift(5:6, 4:6) = 3;
indshift([2,4], 6) = 3;

TE = [TE; zeros(6, 6); repmat(indshift, Xno-3, 1)];

s = 6*(Xno-1)*(Yno-1)*(Zno-2);
TE(s+13:end, :) = TE(s+13:end, :) .* repmat(corval, 1, 6);
TE(s+1:end, :) = TE(s+1:end, :) + repmat(blkE, Xno-2, 1);

clear indshift corval

% Step 8: The last block of the first row on the topest floor
%
e0   = TE(end-5, 1) + 7;
f    = e0 + x1;
g    = e0 + y1;
h    = blkE(5, 4) + 3*(Xno-2);
blkE = ...
[ e0        e0 + 2    e0 + 6    f        f + 2  ( g + x1 ) + 1
  e0        e0 + 4    e0 + 6    f + 1    f + 2    h + 3
  e0 + 1    e0 + 2    e0 + 6    g        g + 4  ( g + x1 ) + 1
  e0 + 1    e0 + 5    e0 + 6    g + 3    g + 4    h + y1_3
  e0 + 3    e0 + 4    e0 + 6    h        h + 2    h + 3
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2    h + y1_3     ];

TE = [TE; blkE];

% Step 9: y-directio shift on the topest floor without the last row block

s  = 6*(Xno-1)*(Yno-1)*(Zno-2);
t1 = s + 6*(Xno-1);
t2 = t1 + 6*(Xno-1);

corval   = reshape(repmat( (2:Yno-3), 6*(Xno-1), 1), [], 1);
indshift = y1*ones(6,6);
indshift(5:6, 4:6) = y1_3;
indshift([2,4], 6) = y1_3;

TE = [TE; repmat(indshift, (Xno-1)*(Yno-3), 1)];
TE(t2+1:end, :) = TE(t2+1:end, :) .* repmat(corval, 1, 6);
TE(t1+1:end, :) = TE(t1+1:end, :) + repmat(TE(s+1:t1, :), Yno-3,1);

clear indshift corval

% Step 10: x-direction shift of the last row on the topest floor.
%         Tetra-edge connections of the first floor without the last block.
%
e0   = TE(end-5, 1) + 10;
f    = e0 + x1;
g    = e0 + y1;
h    = TE(end-1, 6) + 1;
blkE = ...
[ e0        e0 + 2    e0 + 6    f + 1    f + 5  ( g + x1_3 ) + 1
  e0        e0 + 4    e0 + 6    f + 3    f + 5  ( h + x1_3 ) + 1
  e0 + 1    e0 + 2    e0 + 6    g        g + 2  ( g + x1_3 ) + 1
  e0 + 1    e0 + 5    e0 + 6    g + 1    g + 2    h + y1_3
  e0 + 3    e0 + 4    e0 + 6    h        h + 2  ( h + x1_3 ) + 1
  e0 + 3    e0 + 5    e0 + 6    h + 1    h + 2    h + y1_3       ];

corval   = reshape(repmat( (2:Xno-3), 6, 1), [], 1);
indshift = 7*ones(6,6);
indshift(3:6, 4:5) = 3;
indshift([1,2,3,5], 6) = 3;
indshift([4,6], 6) = 1;

s = size(TE, 1);
TE = [TE; zeros(6, 6); repmat(indshift, Xno-3, 1)];
TE(s+13:end, :) = TE(s+13:end, :) .* repmat(corval, 1, 6);
TE(s+1:end, :) = TE(s+1:end, :) + repmat(blkE, Xno-2, 1);

clear corval indshift
% Step 11: Complete the whole tetra-edge connections
e0 = TE(end-5, 1) + 7;
f  = TE(end-3, 4) + x1_3;
g  = TE(end-1, 4) + x1_3;
h  = TE(end,end);
TE = [ TE ;
       [ e0        e0 + 2    e0 + 6    e0 + 7    e0 + 9    f + 3
         e0        e0 + 4    e0 + 6    e0 + 8    e0 + 9    g + 3
         e0 + 1    e0 + 2    e0 + 6     f         f + 2    f + 3
         e0 + 1    e0 + 5    e0 + 6     f + 1     f + 2    h + 1
         e0 + 3    e0 + 4    e0 + 6     g         g + 2    g + 3
         e0 + 3    e0 + 5    e0 + 6     g + 1     g + 2    h + 1 ]       ];
end
%
% ------------------------------------------- END SUBROUTINE :: TE_blkshift